from typing import Optional

from .execption import HuiBiaoException


class Qwen32bAwqException(HuiBiaoException):
    pass


class Qwen32bAwqResponseFormatError(Qwen32bAwqException):
    def __init__(self, msg: str):
        super().__init__(f"模型返回结果错误，报错 {msg}")


class Qwen32bAwqResponseCodeError(Qwen32bAwqException):
    def __init__(self, code: Optional[int]):
        self.code = code
        super().__init__(f"模型处理失败，code={self.code}!")



import os
from typing import Dict, List


class LimitResourceOsConfig:
    """
    有限资源访问限制的环境变量配置
    """

    RES_LIMIT_ENABLE: bool = True
    RES_MAX_NUM: int | Dict[str, int] = {}
    RES_ACQ_TIMES: int | Dict[str, int] = {}
    RES_RETRY_DELAY: int | Dict[str, int] = {}
    REST_WHILE_LIST: List[str] = []

    @classmethod
    def init(cls):
        cls.RES_LIMIT_ENABLE = (
            os.getenv("RES_LIMIT_ENABLE", default="True").lower() == "true"
        )
        cls.RES_MAX_NUM = cls.parse_dict(os.getenv("RES_MAX_NUM", default=16))
        cls.RES_ACQ_TIMES = cls.parse_dict(os.getenv("RES_ACQ_TIMES", default=30))
        cls.RES_RETRY_DELAY = cls.parse_dict(os.getenv("RES_RETRY_DELAY", default=5))
        cls.REST_WHILE_LIST = cls.parse_list(os.getenv("REST_WHILE_LIST", default=""))

    @classmethod
    def get_resource_max_num(cls, resource_name: str):
        if isinstance(cls.RES_MAX_NUM, dict):
            return cls.RES_MAX_NUM.get(resource_name)
        else:
            return cls.RES_MAX_NUM

    @classmethod
    def get_resource_acq_times(cls, resource_name: str):
        if isinstance(cls.RES_ACQ_TIMES, dict):
            return cls.RES_ACQ_TIMES.get(resource_name)
        else:
            return cls.RES_ACQ_TIMES

    @classmethod
    def get_resource_retry_delay(cls, resource_name: str):
        if isinstance(cls.RES_RETRY_DELAY, dict):
            return cls.RES_RETRY_DELAY.get(resource_name)
        else:
            return cls.RES_RETRY_DELAY

    @classmethod
    def parse_list(cls, data: str) -> List[str]:
        return [i for i in data.strip().replace(" ", "").split(",") if i]

    @classmethod
    def is_while_resource(cls, resource_name: str):
        return resource_name in cls.REST_WHILE_LIST

    @classmethod
    def parse_dict(cls, data: int | str) -> Dict[str, int] | int:
        """
        解析资源限制字符串为字典

        Args:
            data: 输入字符串，格式为"name1_value1,name2_value2,..."

        Returns:
            返回LimitResourceName到整数值的映射字典

        例如输入："huizeQwen32bAwq_70,TenderImageOcr_50,TenderLayoutDetect_10"
        则返回：{'huizeQwen32bAwq': 70, 'TenderImageOcr': 50, 'TenderLayoutDetect': 10}
        """
        if isinstance(data, int):
            return data

        try:
            int_data = int(data)
            return int_data
        except Exception:
            pass

        result = {}
        if not data or not isinstance(data, str):
            return result

        # 去除空格并按逗号分割
        items = data.strip().replace(" ", "").split(",")
        for item in items:
            if not item and "_" not in item:
                continue

            try:
                res_name, value = item.split("_")  # 只分割第一个下划线
                num_value = int(value)
                result[res_name] = num_value
            except ValueError:
                continue
        return result


LimitResourceOsConfig.init()



from huibiao_framework.utils.meta_class import OsAttrMeta


class TaskConfig(metaclass=OsAttrMeta):
    TASK_RESOURCE_DIR: str = "/task_resource"


class MinioConfig(metaclass=OsAttrMeta):
    ENDPOINT: str
    AK: str
    SK: str
    BUCKET_NAME: str = "huibiao"
    OSS_SECURE: bool = False


class FfcsConfig(metaclass=OsAttrMeta):
    """
    福富后端
    """

    PROGRESS_URL: str = (
        "http://rubikscube.ffcs:8090/product/tenderprogress/taskProgress"
    )
    CALLBACK_URL: str


class RedisConfig(metaclass=OsAttrMeta):
    REDIS_HOST: str
    REDIS_PORT: int
    REDIS_DB: int = 0
    REDIS_PSWD: str
    REDIS_SENTINEL_PSWD: str
    REDIS_SENTINEL_SERVICE_NAME: str = "mymaster"
    REDIS_MODE: str = "single"  # single | cluster | sentinel


class ModelConfig(metaclass=OsAttrMeta):
    REQUEST_URL: str = "http://vllm-qwen-32b.model.hhht.ctcdn.cn:9080/common/query"
    IMAGE_OCR_TYY_URL: str = (
        "http://tender-document-parser.hhht.ctcdn.cn:9080/image_ocr"
    )
    LAYOUT_DETECTION_TYY_URL: str = (
        "http://tender-document-parser.hhht.ctcdn.cn:9080/image_layout"
    )
    DOCUMENT_PARSER_TYY_URL: str
    CONVERT_TO_PDF_URL: str



import asyncio
import uuid
from typing import Optional, Union, Dict
import redis.asyncio as redis
from loguru import logger
import time
from pydantic import BaseModel

from huibiao_framework.config.limit_resource import LimitResourceOsConfig
from huibiao_framework.execption.limit_resource import (
    LimitResourceAccessTimeOutException,
)
from huibiao_framework.execption.redis import RedisLockAcquireTimeOutException
from huibiao_framework.redis_toolkit import AsyncRedisLock


class LicenceVo(BaseModel):
    access_time: str  # 首次获取时间
    refresh_time: str  # 刷新时间
    user_info: str


class LimitResourceLicence:
    """
    有限资源访问许可证
    """

    HASH_NAME_REFIX = "huibiao:limit-resource-licence"
    LOCK_KEY_PREFIX = "huibiao:limit-resource-licence-lock"

    def __init__(
        self,
        resource_name: str,
        redis_client: Optional[Union[redis.Redis, redis.RedisCluster]],
        resource_max_num: int,  # 资源最大访问数
        acquire_times: int,  # 尝试次数，获取锁失败不会消耗次数
        retry_delay: int,  # 尝试间隔,秒
        user_info: str = None,
    ):
        assert resource_name, "resource_name is required"
        self.resource_name = resource_name

        self.__while_resource = self.is_while(self.resource_name)

        if self.__while_resource:
            return  # 该资源是被设置为白名单

        assert redis_client, "redis_client is required"
        self.__redis_client = redis_client

        self.resource_max_num = (
            resource_max_num
            if resource_max_num is not None and resource_max_num > 0
            else 16
        )
        self.acquire_times = (
            acquire_times if acquire_times is not None and acquire_times > 0 else 20
        )
        self.retry_delay = (
            retry_delay if retry_delay is not None and retry_delay > 0 else 3
        )
        self.user_info = user_info if user_info else ""

        self.hash_name = self.gen_hash_name(self.resource_name)
        self.hash_key = str(uuid.uuid4())
        self.license_vo = None

        self.locker = AsyncRedisLock(
            redis_client=redis_client,
            lock_key=f"{self.LOCK_KEY_PREFIX}:{resource_name}",
            acquire_times=1,
            lock_value=self.hash_key,  # 动态设置
        )

        logger.debug(
            f"Gen res license [{resource_name}][M={self.resource_max_num}][A={self.acquire_times}][D={self.retry_delay}]"
        )

    def desc(self, used_num=None) -> str:
        description = f"[{self.resource_name}][{self.hash_key}]"
        if self.user_info:
            description += f"[{self.user_info}]"
        if used_num is not None:
            description += f"[{used_num}/{self.resource_max_num}]"
        return description

    @classmethod
    def gen_hash_name(cls, resource_name: str):
        return f"{cls.HASH_NAME_REFIX}:{resource_name}"

    async def get_resource_dict(self) -> Dict[str, str]:
        return await self.__redis_client.hgetall(self.hash_name)

    async def resource_current_used_num(self) -> int:
        """
        资源当前被占用的数量
        """
        return len(await self.get_resource_dict())

    def gen_license_vo(self) -> LicenceVo:
        time_now_str = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
        return LicenceVo(
            access_time=time_now_str,
            refresh_time=time_now_str,
            user_info=self.user_info,
        )

    async def acquire(self):
        if self.__while_resource:
            return
        start_time = time.perf_counter()
        i = 1
        while i <= self.acquire_times:
            used_num = await self.resource_current_used_num()
            if used_num >= self.resource_max_num:
                logger.warning(
                    f"{self.desc(used_num)} Resource not enough, try={i}, wait {time.perf_counter() - start_time}s"
                )
                if i < self.acquire_times:
                    await asyncio.sleep(self.retry_delay)
                i += 1
            else:
                is_lock = False
                try:
                    # 资源充足，开始获取锁
                    is_lock = await self.locker.acquire()
                    # 成功获取了锁
                    self.license_vo = self.gen_license_vo()  # 生成vo
                    await self.__redis_client.hset(
                        self.hash_name, self.hash_key, self.license_vo.model_dump_json()
                    )
                    logger.debug(
                        f"{self.desc(used_num)} Get resource success, user_info={self.user_info}, wait {time.perf_counter() - start_time}s, try:{i}"
                    )
                    return  # 结束
                except RedisLockAcquireTimeOutException:
                    logger.warning(f"{self.desc(used_num)} Get resource lock failed. ")
                    if i < self.acquire_times:  # 不是最后一次争抢资源
                        if self.resource_max_num - used_num > 1:
                            await asyncio.sleep(
                                used_num / self.resource_max_num
                            )  # 资源越充足，休眠时间越短
                        else:
                            i += 1
                            await asyncio.sleep(self.retry_delay)
                finally:
                    if is_lock:
                        await self.locker.release()

        waited_time = time.perf_counter() - start_time
        logger.error(
            f"{self.desc()} get resource timeout, wait={waited_time}s, try:{i - 1}"
        )
        assert i - 1 == self.acquire_times
        assert waited_time > self.retry_delay * (self.acquire_times - 1)
        raise LimitResourceAccessTimeOutException(self.resource_name, waited_time)

    async def release(self):
        if self.__while_resource:
            return
        await self.__redis_client.hdel(self.hash_name, self.hash_key)
        logger.debug(f"{self.desc()} Resource release success")

    def is_while(self, resource_name: str):
        # 当前资源是否在白名单中，即无需限制
        return LimitResourceOsConfig.is_while_resource(resource_name)

    async def __aenter__(self):
        """支持 async with 语法"""
        await self.acquire()
        return True

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """支持 async with 语法"""
        await self.release()
        return True



from enum import Enum

from loguru import logger

from .license import LimitResourceLicence
from huibiao_framework.redis_toolkit.client import HuibiaoAsyncRedisClientFactory
from huibiao_framework.config.limit_resource import LimitResourceOsConfig


class DefaultLimitResourceName(str, Enum):
    HUIZE_QWEN_32B_AWQ = "huizeQwen32bAwq"
    TENDER_IMAGE_OCR = "TenderImageOcr"
    TENDER_LAYOUT_DETECT = "TenderLayoutDetect"


class LimitResourceLicenseFactory:
    @classmethod
    def genLicense(
        cls,
        resource_name: str | DefaultLimitResourceName,
        user_info: str = None,
        max_num: int = None,
        acquire: int = None,
        delay: int = None,
    ) -> LimitResourceLicence:
        """
        获取指定资源的访问限制令牌
        """
        resource_name = (
            resource_name.value
            if isinstance(resource_name, DefaultLimitResourceName)
            else resource_name
        )
        max_num = (
            max_num
            if max_num is not None and max_num > 0
            else LimitResourceOsConfig.get_resource_max_num(resource_name)
        )
        acquire = (
            acquire
            if acquire is not None and acquire > 0
            else LimitResourceOsConfig.get_resource_acq_times(resource_name)
        )
        delay = (
            delay
            if delay is not None and delay > 0
            else LimitResourceOsConfig.get_resource_retry_delay(resource_name)
        )
        return LimitResourceLicence(
            resource_name=resource_name,
            redis_client=HuibiaoAsyncRedisClientFactory.get_client(),
            resource_max_num=max_num,
            acquire_times=acquire,
            retry_delay=delay,
            user_info=user_info,
        )

    @classmethod
    def HuiQwen32bAwqLicense(cls, user_info: str = None):
        return cls.genLicense(
            user_info=user_info,
            resource_name=DefaultLimitResourceName.HUIZE_QWEN_32B_AWQ,
        )

    @classmethod
    def TenderImageOcrLicense(cls, user_info: str = None):
        return cls.genLicense(
            user_info=user_info, resource_name=DefaultLimitResourceName.TENDER_IMAGE_OCR
        )

    @classmethod
    def TenderLayoutDetectLicense(cls, user_info: str = None):
        return cls.genLicense(
            user_info=user_info,
            resource_name=DefaultLimitResourceName.TENDER_LAYOUT_DETECT,
        )



from .lock.redis_lock import AsyncRedisLock
from .client import HuibiaoAsyncRedisClientFactory
from huibiao_framework.redis_toolkit.limit_resource import (
    LimitResourceLicence,
    LimitResourceLicenseFactory,
)

__all__ = [
    "AsyncRedisLock",
    "HuibiaoAsyncRedisClientFactory",
    "LimitResourceLicence",
    "LimitResourceLicenseFactory",
]



import os
from typing import Dict, Type, TypeVar

from huibiao_framework.config import TaskConfig
from huibiao_framework.utils.annotation import frozen_attrs
from . import FileOperator
from .extend import JsonOperator

from .result import BatchResult, Result, SingleResult

F = TypeVar("F", bound=FileOperator)
R = TypeVar("R", bound=Result)
D = TypeVar("D")


@frozen_attrs("task_dir", "task_type", "task_id", "resource_dict")
class TaskResult:
    def __init__(self, task_type: str, task_id: str):
        self.task_dir = os.path.join(
            TaskConfig.TASK_RESOURCE_DIR, str(task_type), str(task_id)
        )
        self.task_id = task_id
        self.task_type = task_type
        self.__resource_dict: Dict[str, R] = {}

    def __getitem__(self, item) -> R | None:
        if item in self.result_dict:
            return self.result_dict[item]
        else:
            return None

    @property
    def result_dict(self) -> dict[str, R]:
        return self.__resource_dict

    def addResult(self, name: str, operator: type[F]) -> SingleResult[F, D]:
        """
        name是文件名，不需要带”。***“后缀
        """
        res = SingleResult(
            operator=operator,
            path=os.path.join(self.task_dir, f"{name}.{operator.file_suffix()}"),
        )
        self.result_dict[name] = res
        return res

    def addBatchResult(self, name: str, operator: Type[F]) -> BatchResult[F, D]:
        """
        name是目录名，也是结果名
        """
        res = BatchResult(operator=operator, folder=os.path.join(self.task_dir, name))
        self.result_dict[name] = res
        return operator

    def load(self):
        for _, r in self.result_dict.items():
            r.load()

    def save(self):
        for _, r in self.result_dict.items():
            r.save()

    def addJsonResult(
        self,
        name: str,
    ) -> SingleResult[JsonOperator, Dict]:
        return self.addResult(name=name, operator=JsonOperator)




from pydantic import BaseModel, Field
from typing import Generic, TypeVar, Optional

T = TypeVar("T")


class OcrBaseRespVo(BaseModel, Generic[T]):

    # 响应状态码，0通常表示成功
    code: int = Field(..., description="响应状态码，0表示成功")

    # 响应消息
    message: str = Field("", description="响应状态描述信息")

    # 分析结果数据（泛型类型，可动态指定）
    result: Optional[T] = Field(None, description="分析结果数据，类型由泛型参数指定")



from typing import List, Optional

from pydantic import BaseModel, field_validator
from pydantic_core.core_schema import FieldValidationInfo

from huibiao_framework.execption.vllm import (
    Qwen32bAwqResponseFormatError,
    Qwen32bAwqResponseCodeError,
)


class HuizeQwen32bAwqDto(BaseModel):
    class Message(BaseModel):
        content: str = "今天天气如何"
        role: str = "user"

    Action: str = "NormalChat"
    Messages: List[Message] = [
        Message(),
    ]


class HuizeQwen32bAwqVo(BaseModel):
    class Result(BaseModel):
        Output: Optional[str]
        TokenProbs: Optional[List[float]]

        def remove_think(self) -> str:
            """去掉思考内容"""
            return self.Output.split("</think>")[-1]

        @field_validator("Output")
        def check_output_not_empty(cls, v: str) -> str:
            if v is None or not v.strip():
                raise Qwen32bAwqResponseFormatError("Field 'result.Output' is Empty")
            return v

    code: Optional[int]
    result: Optional[Result]
    message: str

    @field_validator("code")
    def check_code_valid(cls, v: int) -> int:
        if v is None or v != 0:
            # 校验 code 合法性（通常 0 表示成功，非 0 表示错误）
            raise Qwen32bAwqResponseCodeError(v)
        return v

    @field_validator("result")
    def check_result_consistent_with_code(
        cls, v: Optional["Result"], info: FieldValidationInfo
    ) -> Optional["Result"]:
        code = info.data.get("code")
        if code == 0 and v is None:
            # 整体后处理校验：code=0 时 result 必存在
            raise Qwen32bAwqResponseFormatError("Field 'result' is empty!")
        return v



import json
from typing import Dict

from huibiao_framework.result.result import FileOperator


class JsonOperator(FileOperator[Dict]):
    @classmethod
    def file_suffix(cls) -> str:
        return "json"

    @classmethod
    def load(cls, path: str, **kwargs) -> Dict:
        with open(path, "r", encoding="utf-8") as f:
            return json.load(f)

    @classmethod
    def save(cls, data: Dict, path, **kwargs):
        with open(path, "w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=2)
            f.write("\n")



from .execption import HuiBiaoException


class RedisClientException(HuiBiaoException):
    pass


class RedisClientNotInitException(RedisClientException):
    pass


class RedisClientPingException(RedisClientException):
    def __init__(self, ex: Exception = None):
        self.ex = ex
        super().__init__(str(self.ex))


class RedisLockAcquireTimeOutException(RedisClientException):
    def __init__(self, lock_key: str, lock_value: str, waited_time):
        self.lock_key = lock_key
        self.lock_value = lock_value
        self.waited_time = waited_time
        super().__init__(f"无法获取Redis锁，k={lock_key}, v={lock_value}, {waited_time}秒")


import redis.asyncio as redis
from typing import Union, Type
from huibiao_framework.config import RedisConfig
from loguru import logger

from huibiao_framework.execption.redis import (
    RedisClientNotInitException,
    RedisClientPingException,
)


class HuibiaoAsyncRedisClientFactory:
    """
    基于redis.asyncio版本的redis客户端工厂，根据配置创建不同模式的Redis客户端
    """

    __client: Union[redis.Redis, redis.RedisCluster] = None

    @classmethod
    def redis_mode(cls):
        if RedisConfig.REDIS_MODE is None:
            return

    @classmethod
    def init_client(cls) -> Type["HuibiaoAsyncRedisClientFactory"]:
        if cls.__client is not None:
            return cls

        logger.debug(f"初始化Redis客户端，模式：{RedisConfig.REDIS_MODE}")

        if RedisConfig.REDIS_MODE.lower() == "single":
            pool = redis.ConnectionPool(
                host=RedisConfig.REDIS_HOST,
                port=RedisConfig.REDIS_PORT,
                db=RedisConfig.REDIS_DB,
                password=RedisConfig.REDIS_PSWD,
            )
            cls.__client = redis.Redis(connection_pool=pool)
            logger.debug(f"Redis客户端初始化完成，连接信息：{pool.connection_kwargs}")

        elif RedisConfig.REDIS_MODE.lower() == "sentinel":
            sentinel = redis.sentinel.Sentinel(
                sentinels=[(RedisConfig.REDIS_HOST, RedisConfig.REDIS_PORT)],
                password=RedisConfig.REDIS_PSWD,
                sentinel_kwargs={"password": RedisConfig.REDIS_SENTINEL_PSWD},
            )
            cls.__client = sentinel.master_for(
                service_name=RedisConfig.REDIS_SENTINEL_SERVICE_NAME,
                db=RedisConfig.REDIS_DB,
            )
            logger.debug(
                f"Redis客户端初始化完成，连接信息：{sentinel.connection_kwargs}"
            )

        elif RedisConfig.REDIS_MODE.lower() == "cluster":
            cls.__client = redis.RedisCluster(
                host=RedisConfig.REDIS_HOST,
                port=RedisConfig.REDIS_PORT,
                password=RedisConfig.REDIS_PSWD,
            )
            logger.debug(
                f"Redis客户端初始化完成，连接信息：{cls.__client.connection_pool.connection_kwargs}"
            )
        else:
            logger.debug(f"Redis模式错误 {RedisConfig.REDIS_MODE}")

        return cls

    @classmethod
    async def ping(cls):
        if cls.__client is not None:
            try:
                await cls.__client.ping()
                logger.debug("Redis客户端连接成功")
            except Exception as e:
                logger.error(f"Redis客户端连接失败：{str(e)}")
                raise RedisClientPingException(e)
        else:
            logger.error("Redis客户端未初始化")
            raise RedisClientNotInitException()

    @classmethod
    def get_client(cls) -> Union[redis.Redis, redis.RedisCluster]:
        cls.init_client()
        return cls.__client



import asyncio
import time
from typing import Union, Optional
from loguru import logger
import redis.asyncio as redis

from huibiao_framework.execption.redis import RedisLockAcquireTimeOutException
from huibiao_framework.redis_toolkit.client import HuibiaoAsyncRedisClientFactory
from huibiao_framework.utils.annotation import frozen_attrs


@frozen_attrs("lock_key", "lock_value")
class AsyncRedisLock:
    def __init__(
        self,
        lock_key: str,
        lock_value: str,
        *,
        redis_client: Optional[Union[redis.Redis, redis.RedisCluster]] = None,
        expire_time: int = 10,
        acquire_times: int = 30,
        retry_delay: int = 5,
    ):
        self.client = redis_client
        self.lock_key = lock_key
        self.lock_value = lock_value
        self.expire_time = expire_time  # 锁超时时间，秒
        self.acquire_times = acquire_times  # 尝试次数
        self.retry_delay = retry_delay  # 重试间隔，秒

    async def init(self) -> "AsyncRedisLock":
        if self.client is None:
            self.client = await HuibiaoAsyncRedisClientFactory.get_client()
        return self

    async def acquire(self) -> bool:
        """使用 SETNX 原子性获取锁"""
        start_time = time.perf_counter()

        for i in range(self.acquire_times):
            # SET key value NX EX expire 等同于 SETNX + EXPIRE 原子操作
            acquired = await self.client.set(
                self.lock_key,
                self.lock_value,
                nx=True,  # 仅在键不存在时设置（SETNX 效果）
                ex=self.expire_time,  # 设置过期时间，防止死锁 秒
            )
            if acquired:
                logger.debug(
                    f"获取锁成功：{self.lock_key} -> {self.lock_value}, 等待{time.perf_counter() - start_time:.6f}s, 尝试{i + 1}次"
                )
                return True
            if i < self.acquire_times - 1:
                await asyncio.sleep(self.retry_delay)

        logger.warning(
            f"获取锁失败：{self.lock_key} -> {self.lock_value}, 尝试{self.acquire_times}次"
        )
        raise RedisLockAcquireTimeOutException(
            self.lock_key, self.lock_value, time.perf_counter() - start_time
        )

    async def release(self):
        script = """
        if redis.call("GET", KEYS[1]) == ARGV[1] then
            return redis.call("DEL", KEYS[1])
        else
            return 0
        end
        """
        result = await self.client.eval(script, 1, self.lock_key, self.lock_value)
        logger.debug(
            f"释放锁成功：{self.lock_key} -> {self.lock_value}, result={result}"
        )

    async def __aenter__(self):
        """支持 async with 语法"""
        return await self.acquire() if self.lock_key else None

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """支持 async with 语法"""
        await self.release()



from .execption import HuiBiaoException


class FfcsClientError(HuiBiaoException):
    pass


class FfcsSendProgressError(FfcsClientError):
    def __init__(self, reason, e: Exception = None, reqid: str = None):
        self.reason = reason
        self.e = e
        self.reqid = reqid
        super().__init__(f"发送进度失败,reqid={self.reqid},原因：{self.reason}")
